# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness: https://github.com/VarIr/scikit-hubness/

from typing import Tuple, TypeVar

import numpy as np
from sklearn.metrics import euclidean_distances
from sklearn.utils.extmath import row_norms
from sklearn.utils.validation import check_is_fitted

from .base import HubnessReduction

T = TypeVar("T")

_DESIRED_P_VALUE = 2
_MINIMUM_DIST = 0.0

try:
    import torch
except ImportError:
    torch = None


class DisSimLocal(HubnessReduction):
    """Hubness reduction with DisSimLocal.

    Uses the formula presented in [1]_.

    Parameters
    ----------
    squared: bool, default = True
        DisSimLocal operates on squared Euclidean distances.
        If True, return (quasi) squared Euclidean distances;
        if False, return (quasi) Eucldean distances.

    References
    ----------
    .. [1] Hara K, Suzuki I, Kobayashi K, Fukumizu K, Radovanović M (2016)
           Flattening the density gradient for eliminating spatial centrality to reduce hubness.
           In: Proceedings of the 30th AAAI conference on artificial intelligence, pp 1659-1665.
           https://www.aaai.org/ocs/index.php/AAAI/AAAI16/paper/viewPaper/12055
    """

    def __init__(self, squared: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.squared = squared
        if self.nn_algo.metric in ["euclidean", "minkowski"]:
            self.squared = False
            if hasattr(self.nn_algo, "p") and self.nn_algo.p != _DESIRED_P_VALUE:
                raise ValueError(
                    "DisSimLocal only supports squared Euclidean distances. If"
                    " the provided NNAlgorithm has a `p` parameter it must be"
                    f" set to p=2. Now it is p={self.nn_algo.p}"
                )
        elif self.nn_algo.metric in ["sqeuclidean"]:
            self.squared = True
        else:
            raise ValueError(
                "DisSimLocal only supports squared Euclidean distances, not"
                f" metric={self.nn_algo.metric}."
            )

    def __repr__(self):
        return f"{self.__class__.__name__}(squared = {self.squared})"

    def _fit(
        self,
        neigh_dist,
        neigh_ind,
        source,
        target,
    ) -> "DisSimLocal":
        """Fit the model using target, neigh_dist, and neigh_ind as training data.

        Parameters
        ----------
        neigh_dist: shape (n_samples, n_neighbors)
            Distance matrix of training objects (rows) against their
            individual k nearest neighbors (colums).
        neigh_ind: shape (n_samples, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist.
        source: shape (n_samples, n_features)
            source embedding, where n_samples is the number of vectors,
            and n_features their dimensionality (number of features).
        target: shape (n_samples, n_features)
            Target embedding, where n_samples is the number of vectors,
            and n_features their dimensionality (number of features).

        Returns
        -------
        DisSimLocal
            Fitted DisSimLocal
        """
        # Calculate local neighborhood centroids among the training points
        knn = neigh_ind
        centroids = source[knn].mean(axis=1)
        if self._use_torch:
            # see https://github.com/scikit-learn/scikit-learn/blob/main/sklearn/utils/extmath.py#L87C21-L87C48
            X = target - centroids
            dist_to_cent = torch.einsum("ij,ij->i", X, X)
        else:
            dist_to_cent = row_norms(target - centroids, squared=True)

        self.source_ = source
        self.target_ = target
        self.target_centroids_ = centroids
        self.target_dist_to_centroids_ = dist_to_cent
        return self

    def transform(
        self,
        neigh_dist,
        neigh_ind,
        query,
    ) -> Tuple[T, T]:
        """Transform distance between test and training data with DisSimLocal.

        Parameters
        ----------
        neigh_dist: shape (n_query, n_neighbors)
            Distance matrix of test objects (rows) against their individual
            k nearest neighbors among the training data (columns).
        neigh_ind: shape (n_query, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist
        query: shape (n_query, n_features)
            Query entities that were used to obtain neighbors
            If none is provided use source that was provided in fit step

        Returns
        -------
        hub_reduced_dist, neigh_ind
            DisSimLocal distances, and corresponding neighbor indices

        Notes
        -----
        The returned distances are NOT sorted! If you use this class directly,
        you will need to sort the returned matrices according to hub_reduced_dist.
        """
        check_is_fitted(
            self,
            ["target_", "target_centroids_", "target_dist_to_centroids_"],
        )
        # Calculate local neighborhood centroids for source objects among target objects
        if self._use_torch:
            # pairwise squared euclidean distance between each query vector and knn
            # unsqueeze to enable batching
            hub_reduced_dist = (
                torch.cdist(torch.unsqueeze(query, 1), self.target_[neigh_ind])
                .pow(2)
                .squeeze()
            )
        else:
            hub_reduced_dist = np.empty_like(neigh_dist)
            for i, ind in enumerate(neigh_ind):
                hub_reduced_dist[i, :] = euclidean_distances(
                    query[i].reshape(1, -1), self.target_[ind], squared=True
                )

        centroids = self.target_[neigh_ind].mean(axis=1)
        source_minus_centroids = query - centroids
        source_minus_centroids **= 2
        source_dist_to_centroids = source_minus_centroids.sum(axis=1)
        target_dist_to_centroids = self.target_dist_to_centroids_[neigh_ind]

        hub_reduced_dist -= source_dist_to_centroids.reshape(-1, 1)
        hub_reduced_dist -= target_dist_to_centroids

        # DisSimLocal can yield negative dissimilarities, which can cause problems with
        # certain scikit-learn routines (e.g. in metric='precomputed' usages).
        # We, therefore, shift dissimilarities to non-negative values, if necessary.
        min_dist = hub_reduced_dist.min()
        if min_dist < _MINIMUM_DIST:
            hub_reduced_dist += -min_dist

        # Return Euclidean or squared Euclidean distances?
        if not self.squared:
            hub_reduced_dist **= 1 / 2

        # Return the hubness reduced distances
        # These must be sorted downstream
        return hub_reduced_dist, neigh_ind



from typing import Tuple, TypeVar

from sklearn.utils.validation import check_is_fitted

from .base import HubnessReduction

T = TypeVar("T")


class CSLS(HubnessReduction):
    """Hubness reduction with Cross-domain similarity local scaling.

    Uses the formula presented in [1]_.

    References
    ----------
    .. [1] Lample, G., Conneau, A., Ranzato, M., Denoyer, L., & Jégou, H. (2018)
           Word translation without parallel data
           In: 6th International Conference on Learning Representations,
           ICLR 2018 - Conference Track Proceedings.
           https://openreview.net/forum?id=H196sainb
    """

    def __repr__(self):
        return f"{self.__class__.__name__}(verbose = {self.verbose})"

    def _fit(
        self,
        neigh_dist,
        neigh_ind,
        source=None,
        target=None,
    ) -> "CSLS":
        """Fit the model using target, neigh_dist, and neigh_ind as training data.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_samples, n_neighbors)
            Distance matrix of training objects (rows) against their
            individual k nearest neighbors (colums).
        neigh_ind: np.ndarray, shape (n_samples, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist.
        source
            ignored
        target
            ignored

        Returns
        -------
        CSLS
            Fitted CSLS
        """
        self.r_dist_train_ = neigh_dist
        self.r_ind_train_ = neigh_ind
        return self

    def transform(
        self,
        neigh_dist,
        neigh_ind,
        query,
    ) -> Tuple[T, T]:
        """Transform distance between test and training data with CSLS.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_query, n_neighbors)
            Distance matrix of test objects (rows) against their individual
            k nearest neighbors among the training data (columns).
        neigh_ind: np.ndarray, shape (n_query, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist
        query
            Ignored

        Returns
        -------
        hub_reduced_dist, neigh_ind
            CSLS distances, and corresponding neighbor indices

        Notes
        -----
        The returned distances are NOT sorted! If you use this class directly,
        you will need to sort the returned matrices according to hub_reduced_dist.
        """
        check_is_fitted(self, "r_dist_train_")

        # Find average distances to the k nearest neighbors
        r_dist_test = neigh_dist

        r_train = self.r_dist_train_.mean(axis=1)
        r_test = r_dist_test.mean(axis=1).reshape(-1, 1)

        hub_reduced_dist = 2 * neigh_dist - r_test - r_train[neigh_ind]
        # Return the hubness reduced distances
        # These must be sorted downstream
        return hub_reduced_dist, neigh_ind



import json
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, TypeVar, Union, overload

import numpy as np
from class_resolver import HintOrType

from kiez.hubness_reduction import (
    hubness_reduction_resolver,
)
from kiez.hubness_reduction.base import HubnessReduction
from kiez.neighbors import NNAlgorithm, nn_algorithm_resolver
from kiez.neighbors.util import available_nn_algorithms

T = TypeVar("T")


class Kiez:
    """Performs hubness reduced nearest neighbor search for entity alignment.

    Use the given algorithm to :meth:`fit` the data and calculate the
    :meth:`kneighbors`.

    Parameters
    ----------
    n_neighbors : int, default=5
        number of nearest neighbors used in search
    algorithm : :obj:`~kiez.neighbors.NNAlgorithm`, default = None
        initialised `NNAlgorithm` object that will be used for neighbor search
        If no algorithm is provided :obj:`~kiez.neighbors.Faiss` is used if available else
        :obj:`~kiez.neighbors.SklearnNN` is used with default values
    algorithm_kwargs :
        A dictionary of keyword arguments to pass to the :obj:`~kiez.neighbors.NNAlgorithm`
        if given as a class in the ``algorithm`` argument.

    hubness :
        Either an instance of a :obj:`~kiez.hubness_reduction.base.HubnessReduction`,
        the class for a :obj:`~kiez.hubness_reduction.base.HubnessReduction` that should
        be instantiated, the name of the hubness reduction method, or if None, defaults
        to no hubness reduction.
    hubness_kwargs :
        A dictionary of keyword arguments to pass to the :obj:`~kiez.hubness_reduction.base.HubnessReduction`
        if given as a class in the ``hubness`` argument.

    Examples
    --------
    >>> from kiez import Kiez
    >>> import numpy as np
    >>> # create example data
    >>> rng = np.random.RandomState(0)
    >>> source = rng.rand(100,50)
    >>> target = rng.rand(100,50)
    >>> # fit and get neighbors
    >>> k_inst = Kiez()
    >>> k_inst.fit(source, target)
    >>> nn_dist, nn_ind = k_inst.kneighbors()

    Using a specific algorithm and hubness reduction

    >>> from kiez import Kiez
    >>> import numpy as np
    >>> # create example data
    >>> rng = np.random.RandomState(0)
    >>> source = rng.rand(100,50)
    >>> target = rng.rand(100,50)
    >>> # prepare algorithm and hubness reduction
    >>> from kiez.neighbors import NMSLIB
    >>> hnsw = NMSLIB(n_candidates=10)
    >>> from kiez.hubness_reduction import CSLS
    >>> hr = CSLS()
    >>> # fit and get neighbors
    >>> k_inst = Kiez(n_neighbors=5, algorithm=hnsw, hubness=hr)
    >>> k_inst.fit(source, target)
    >>> nn_dist, nn_ind = k_inst.kneighbors()

    NN and hubness algorithms can also be supplied via string:

    >>> k_inst = Kiez(algorithm="SklearnNN", hubness="CSLS")

    You can investigate which NN algos are installed and which hubness methods are implemented with:

    >>> Kiez.show_hubness_options()
    >>> Kiez.show_algorithm_options()

    Beginning with version 0.5.0 torch can be used, when using `Faiss` as NN algorithm:

    >>> from kiez import Kiez
    >>> import torch
    >>> source = torch.randn((100,10))
    >>> target = torch.randn((200,10))
    >>> k_inst = Kiez(algorithm="Faiss", hubness="CSLS")
    >>> k_inst.fit(source, target)
    >>> nn_dist, nn_ind = k_inst.kneighbors()

    You can also utilize tensors and NN calculations on the GPU:

    >>> k_inst = Kiez(algorithm="Faiss", algorithm_kwargs={"use_gpu":True}, hubness="CSLS")
    >>> k_inst.fit(source.cuda(), target.cuda())
    >>> nn_dist, nn_ind = k_inst.kneighbors()

    You can also initalize Kiez via a json file

    >>> kiez = Kiez.from_path("tests/example_conf.json")
    """

    def __init__(
        self,
        n_candidates: int = 10,
        algorithm: HintOrType[NNAlgorithm] = None,
        algorithm_kwargs: Optional[Dict[str, Any]] = None,
        hubness: HintOrType[HubnessReduction] = None,
        hubness_kwargs: Optional[Dict[str, Any]] = None,
    ):
        if not np.issubdtype(type(n_candidates), np.integer):
            raise TypeError(
                f"n_neighbors does not take {type(n_candidates)} value, enter"
                " integer value"
            )

        if n_candidates <= 0:
            raise ValueError(f"Expected n_candidates > 0. Got {n_candidates}")
        if algorithm_kwargs is None:
            algorithm_kwargs = {"n_candidates": n_candidates}
        elif "n_candidates" not in algorithm_kwargs:
            algorithm_kwargs["n_candidates"] = n_candidates
        if algorithm is None:
            try:
                algorithm = nn_algorithm_resolver.make("Faiss", algorithm_kwargs)
            except ImportError:
                algorithm = nn_algorithm_resolver.make("SklearnNN", algorithm_kwargs)
        else:
            algorithm = nn_algorithm_resolver.make(algorithm, algorithm_kwargs)
        assert algorithm
        if hubness_kwargs is None:
            hubness_kwargs = {}
        hubness_kwargs["nn_algo"] = algorithm
        self.hubness = hubness_reduction_resolver.make(hubness, hubness_kwargs)

    @staticmethod
    def show_algorithm_options() -> List[str]:
        return available_nn_algorithms(as_string=True)

    @staticmethod
    def show_hubness_options() -> List[str]:
        return list(hubness_reduction_resolver.options)

    @property
    def algorithm(self):
        return self.hubness.nn_algo

    @algorithm.setter
    def algorithm(self, value):
        self.hubness.nn_algo = value

    def __repr__(self):
        return (
            f"Kiez(algorithm: {self.algorithm},"
            f" hubness: {self.hubness})"
            f" {self.algorithm._describe_source_target_fitted()}"
        )

    @classmethod
    def from_path(cls, path: Union[str, Path]) -> "Kiez":
        """Load a Kiez instance from configuration in a JSON file, based on its path."""
        with open(path) as file:
            return cls(**json.load(file))

    def fit(self, source: T, target: Optional[T] = None) -> "Kiez":
        """Fits the algorithm and hubness reduction method.

        Parameters
        ----------
        source : matrix of shape (n_samples, n_features)
            embeddings of source entities
        target : matrix of shape (m_samples, n_features)
            embeddings of target entities. If none given, uses the source.

        Returns
        -------
        Kiez
            Fitted kiez instance
        """
        self.hubness.fit(source, target)
        return self

    @overload
    def kneighbors(
        self,
        k: Optional[int] = None,
        return_distance: Literal[True] = True,
    ) -> Tuple[T, T]:
        ...

    @overload
    def kneighbors(
        self,
        k: Optional[int] = None,
        return_distance: Literal[False] = False,
    ) -> Any:
        ...

    def kneighbors(
        self,
        k: Optional[int] = None,
        return_distance: bool = True,
    ) -> Union[T, Tuple[T, T]]:
        """Retrieve the k-nearest neighbors using the supplied nearest neighbor algorithm and hubness reduction method.

        Parameters
        ----------
        k : Optional[int], default = None
            k-nearest neighbors, if None is set to number of n_candidates
        return_distance : bool, default = True
            Whether to return distances
            If `False` only indices are returned

        Returns
        -------
        neigh_dist : ndarray of shape (n_queries, n_neighbors)
            Array representing the distance between source and target entities
            only present if return_distance=True.
        neigh_ind : ndarray of shape (n_queries, n_neighbors)
            Indices of the nearest points in the population matrix.
        """
        hubness_reduced_query_dist, query_ind = self.hubness.kneighbors(k)

        if return_distance:
            result = hubness_reduced_query_dist, query_ind
        else:
            result = query_ind
        return result



import logging
from types import MappingProxyType
from typing import Optional

import numpy as np

from kiez.neighbors.neighbor_algorithm_base import NNAlgorithm

try:
    import faiss
except ImportError:  # pragma: no cover
    faiss = None

try:
    import torch  # noqa: I001
    import faiss.contrib.torch_utils
except ImportError:  # pragma: no cover
    torch = None


class Faiss(NNAlgorithm):
    """Wrapper for `faiss` library.

    Faiss implements a number of (A)NN algorithms and enables the use of GPUs.

    Parameters
    ----------
    n_candidates: int, default = 5
        number of nearest neighbors used in search
    metric: str, default = 'l2'
        distance measure used in search
        possible measures are found in :obj:`Faiss.valid_metrics`
        Euclidean is the same as l2, expect for taking the sqrt of the result
    index_key: str, default = None
        index name to use
        If none is provided will determine the best automatically
        Else will use it as input for :meth:`faiss.index_factory`
    index_param: str, default = None
        Hyperparameter string for the indexing algorithm
        See https://github.com/facebookresearch/faiss/wiki/Index-IO,-cloning-and-hyper-parameter-tuning#auto-tuning-the-runtime-parameters for info
    use_gpu: bool
        If true uses all available gpus

    Examples
    --------
    >>> import numpy as np
    >>> from kiez import Kiez
    >>> source = np.random.rand(1000, 512)
    >>> target = np.random.rand(100, 512)
    >>> k_inst = Kiez(algorithm="Faiss")
    >>> k_inst.fit(source, target)

    >>> k_inst = Kiez(algorithm="Faiss",algorithm_kwargs={"metric":"euclidean","index_key":"Flat"})

    supply hyperparameters for indexing algorithm

    >>> k_inst = Kiez(algorithm="Faiss",algorithm_kwargs={"index_key":"HNSW32","index_param":"efSearch=16383"})

    Notes
    -----
    For details about configuring faiss consult their wiki: https://github.com/facebookresearch/faiss/wiki
    """

    if torch:
        _ALLOWED_INPUT_TYPES = (np.ndarray, torch.Tensor)

    _METRIC_MAP = MappingProxyType({})  # type: ignore [var-annotated]
    if faiss:
        _METRIC_MAP = MappingProxyType(
            {
                "euclidean": faiss.METRIC_L2,
                "l2": faiss.METRIC_L2,
                "l1": faiss.METRIC_L1,
                "ip": faiss.METRIC_INNER_PRODUCT,
                "innerproduct": faiss.METRIC_INNER_PRODUCT,
                "cosine": faiss.METRIC_INNER_PRODUCT,
                "braycurtis": faiss.METRIC_BrayCurtis,
                "canberra": faiss.METRIC_Canberra,
                "jensenshannon": faiss.METRIC_JensenShannon,
                "chebyshev": faiss.METRIC_Linf,
                "linf": faiss.METRIC_Linf,
            }
        )

    valid_metrics = tuple(_METRIC_MAP.keys())

    def __init__(
        self,
        n_candidates: int = 5,
        metric: str = "l2",
        index_key: str = "Flat",
        index_param: Optional[str] = None,
        use_gpu: bool = False,
        verbose: int = logging.WARNING,
    ):
        if faiss is None:  # pragma: no cover
            raise ImportError(
                "Please install the `faiss` package, before using this class.\nSee here"
                " for installation instructions:"
                " https://github.com/facebookresearch/faiss/blob/main/INSTALL.md"
            )
        if metric not in self.__class__.valid_metrics:
            raise ValueError(
                f"Unknown metric {metric}, please use one of {self.valid_metrics}"
            )
        super().__init__(n_candidates=n_candidates, metric=metric, n_jobs=None)
        self.index_key = index_key
        self.index_param = index_param
        self.use_gpu = use_gpu
        self.verbose = verbose
        self._faiss_metric = self.__class__._METRIC_MAP[metric]

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(n_candidates={self.n_candidates},"
            + f"metric={self.metric},"
            + f"index_key={self.index_key},"
            + f"index_param={{{self.index_param}}},"
            + f"use_gpu={self.use_gpu})"
        )

    def _normalize_if_needed(self, vec):
        if self.metric == "cosine":
            # see https://github.com/facebookresearch/faiss/wiki/MetricType-and-distances#how-can-i-index-vectors-for-cosine-similarity
            if torch and isinstance(vec, torch.Tensor):
                return torch.nn.functional.normalize(vec)
            if vec.dtype != "float32":
                vec = vec.astype("float32")
            # normalizes in-place!
            faiss.normalize_L2(vec)
        return vec

    def _fit(self, data, is_source: bool):
        dim = data.shape[1]
        index = faiss.index_factory(dim, self.index_key, self._faiss_metric)
        params = faiss.ParameterSpace()
        if self.use_gpu:
            index = faiss.index_cpu_to_all_gpus(index)
            params = faiss.GpuParameterSpace()
        if self.index_param is not None:
            params.set_index_parameters(index, self.index_param)
        data = self._normalize_if_needed(data)
        index.add(data)
        return index

    def _kneighbors(self, query, k, index, return_distance, is_self_querying):
        if is_self_querying:
            query = self.source_
        query = self._normalize_if_needed(query)
        dist, ind = index.search(query, k)
        if return_distance:
            if self.metric == "euclidean":
                if torch and isinstance(dist, torch.Tensor):
                    dist = torch.sqrt(dist)
                else:
                    dist = np.sqrt(dist)
            return dist, ind
        return ind



from typing import List, Literal, Type, Union, overload

from kiez.neighbors import NNAlgorithm, nn_algorithm_resolver


@overload
def available_nn_algorithms(as_string: Literal[True]) -> List[str]:
    ...


@overload
def available_nn_algorithms(
    as_string: Literal[False] = False,
) -> List[Type[NNAlgorithm]]:
    ...


def available_nn_algorithms(
    as_string: bool = False,
) -> Union[List[str], List[Type[NNAlgorithm]]]:
    """Get available (approximate) nearest neighbor algorithms.

    Returns
    -------
    algorithms: List[Type[NNAlgorithm]]
        A tuple of available algorithms
    """
    possible = ["NMSLIB", "NNG", "Annoy", "Faiss", "SklearnNN"]
    available = []
    for ann in possible:
        try:
            nn_algorithm_resolver.make(ann)
            if as_string:
                available.append(ann.lower())
            else:
                available.append(nn_algorithm_resolver.lookup(ann))
        except ImportError:  # noqa: PERF203
            pass
    return available



import warnings
from abc import ABC, abstractmethod
from typing import Optional, Tuple, TypeVar

import numpy as np

from ..neighbors import NNAlgorithm

try:
    import torch
except ImportError:
    torch = None

T = TypeVar("T")


class HubnessReduction(ABC):
    """Base class for hubness reduction."""

    def __init__(self, nn_algo: NNAlgorithm, verbose: int = 0, **kwargs):
        self.nn_algo = nn_algo
        self.verbose = verbose
        self._use_torch = False
        if nn_algo.n_candidates == 1:
            raise ValueError(
                "Cannot perform hubness reduction with a single candidate per query!"
            )

    @abstractmethod
    def _fit(self, neigh_dist: T, neigh_ind: T, source: T, target: T):
        pass  # pragma: no cover

    def fit(self, source: T, target: Optional[T] = None):
        self.nn_algo.fit(source, target)
        if target is None:
            target = source
        neigh_dist_t_to_s, neigh_ind_t_to_s = self.nn_algo.kneighbors(
            k=self.nn_algo.n_candidates,
            query=target,
            s_to_t=False,
            return_distance=True,
        )
        if torch and isinstance(neigh_dist_t_to_s, torch.Tensor):
            self._use_torch = True
        self._fit(
            neigh_dist_t_to_s,
            neigh_ind_t_to_s,
            source,
            target,
        )

    @abstractmethod
    def transform(self, neigh_dist, neigh_ind, query) -> Tuple[T, T]:
        pass  # pragma: no cover

    def _set_k_if_needed(self, k: Optional[int] = None) -> int:
        if k is None:
            warnings.warn(
                f"No k supplied, setting to n_candidates = {self.nn_algo.n_candidates}",
                stacklevel=2,
            )
            return self.nn_algo.n_candidates
        if k > self.nn_algo.n_candidates:
            warnings.warn(
                "k > n_candidates supplied! Setting to n_candidates ="
                f" {self.nn_algo.n_candidates}",
                stacklevel=2,
            )
            return self.nn_algo.n_candidates
        return k

    @staticmethod
    def _sort(hubness_reduced_query_dist, query_ind, n_neighbors: int) -> Tuple[T, T]:
        if torch and isinstance(hubness_reduced_query_dist, torch.Tensor):
            mask = torch.argsort(hubness_reduced_query_dist)[:, :n_neighbors]
            hubness_reduced_query_dist = torch.take_along_dim(
                hubness_reduced_query_dist, mask, dim=1
            )
            query_ind = torch.take_along_dim(query_ind, mask, dim=1)
        else:
            kth = np.arange(n_neighbors)
            mask = np.argpartition(hubness_reduced_query_dist, kth=kth)[:, :n_neighbors]
            hubness_reduced_query_dist = np.take_along_axis(
                hubness_reduced_query_dist, mask, axis=1
            )
            query_ind = np.take_along_axis(query_ind, mask, axis=1)
        return hubness_reduced_query_dist, query_ind

    def kneighbors(self, k: Optional[int] = None) -> Tuple[T, T]:
        n_neighbors = self._set_k_if_needed(k)
        # First obtain candidate neighbors
        query_dist, query_ind = self.nn_algo.kneighbors(
            query=None, k=self.nn_algo.n_candidates, return_distance=True
        )

        # Second, reduce hubness
        hubness_reduced_query_dist, query_ind = self.transform(
            query_dist,
            query_ind,
            self.nn_algo.source_,
        )
        # Third, sort hubness reduced candidate neighbors to get the final k neighbors
        return HubnessReduction._sort(
            hubness_reduced_query_dist, query_ind, n_neighbors
        )


class NoHubnessReduction(HubnessReduction):
    """Base class for hubness reduction."""

    def _fit(self, neigh_dist, neigh_ind, source, target):
        pass  # pragma: no cover

    def fit(self, source, target=None):
        self.nn_algo.fit(source, target, only_fit_target=True)

    def transform(self, neigh_dist, neigh_ind, query) -> Tuple[T, T]:
        return neigh_dist, neigh_ind

    def kneighbors(self, k: Optional[int] = None) -> Tuple[T, T]:
        n_neighbors = self._set_k_if_needed(k)
        return self.nn_algo.kneighbors(query=None, k=n_neighbors, return_distance=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kiez',
 'kiez.analysis',
 'kiez.evaluate',
 'kiez.hubness_reduction',
 'kiez.io',
 'kiez.neighbors',
 'kiez.neighbors.approximate',
 'kiez.neighbors.exact']

package_data = \
{'': ['*']}

modules = \
['py']
install_requires = \
['class-resolver>=0.3',
 'joblib>=1.2.0,<2.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.1.5,<2.0.0',
 'scikit-learn>=0.24.1,<2.0.0',
 'scipy>=1.3.2,<2.0.0',
 'tqdm>=4.62.3,<5.0.0']

extras_require = \
{'annoy': ['annoy>=1.17.0,<2.0.0'],
 'docs': ['Sphinx>=5.0.0,<6.0.0', 'insegel>=1.3.1,<2.0.0'],
 'ngt': ['ngt>=1.8,<2.0'],
 'nmslib': ['nmslib>=2.1.1,<3.0.0']}

setup_kwargs = {
    'name': 'kiez',
    'version': '0.5.0',
    'description': 'Hubness reduced nearest neighbor search for entity alignment with knowledge graph embeddings',
    'long_description': '<p align="center">\n<img src="https://github.com/dobraczka/kiez/raw/main/docs/kiezlogo.png" alt="kiez logo", width=200/>\n</p>\n\n<h2 align="center"> <a href="https://dbs.uni-leipzig.de/file/KIEZ_KEOD_2021_Obraczka_Rahm.pdf">kiez</a></h2>\n\n<p align="center">\n<a href="https://github.com/dobraczka/kiez/actions/workflows/main.yml"><img alt="Actions Status" src="https://github.com/dobraczka/kiez/actions/workflows/main.yml/badge.svg?branch=main"></a>\n<a href=\'https://kiez.readthedocs.io/en/latest/?badge=latest\'><img src=\'https://readthedocs.org/projects/kiez/badge/?version=latest\' alt=\'Documentation Status\' /></a>\n<a href="https://pypi.org/project/kiez"/><img alt="Stable python versions" src="https://img.shields.io/pypi/pyversions/kiez"></a>\n<a href="https://github.com/dobraczka/kiez/blob/main/LICENSE"><img alt="License BSD3 - Clause" src="https://img.shields.io/badge/license-BSD--3--Clause-blue"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\nA Python library for hubness reduced nearest neighbor search for the task of entity alignment with knowledge graph embeddings. The term kiez is a [german word](https://en.wikipedia.org/wiki/Kiez) that refers to a city neighborhood.\n\n## Hubness Reduction\nHubness is a phenomenon that arises in high-dimensional data and describes the fact that a couple of entities are nearest neighbors (NN) of many other entities, while a lot of entities are NN to no one.\nFor entity alignment with knowledge graph embeddings we rely on NN search. Hubness therefore is detrimental to our matching results.\nThis library is intended to make hubness reduction techniques available to data integration projects that rely on (knowledge graph) embeddings in their alignment process. Furthermore kiez incorporates several approximate nearest neighbor (ANN) libraries, to pair the speed advantage of approximate neighbor search with increased accuracy of hubness reduction.\n\n## Installation\nYou can install kiez via pip:\n``` bash\npip install kiez\n```\n\nIf you have a GPU you can make kiez faster by installing [faiss](https://github.com/facebookresearch/faiss) (if you do not already have it in your environment):\n\n``` bash\nconda env create -n kiez-faiss python=3.10\nconda activate kiez-faiss\nconda install -c pytorch -c nvidia faiss-gpu=1.7.4 mkl=2021 blas=1.0=mkl\npip install kiez\n```\n\nFor more information see their [installation instructions](https://github.com/facebookresearch/faiss/blob/main/INSTALL.md).\n\nYou can also get other specific libraries with e.g.:\n\n``` bash\n  pip install kiez[nmslib]\n```\n\n## Usage\nSimple nearest neighbor search for source entities in target space:\n``` python\nfrom kiez import Kiez\nimport numpy as np\n# create example data\nrng = np.random.RandomState(0)\nsource = rng.rand(100,50)\ntarget = rng.rand(100,50)\n# fit and get neighbors\nk_inst = Kiez()\nk_inst.fit(source, target)\nnn_dist, nn_ind = k_inst.kneighbors()\n```\nUsing (A)NN libraries and hubness reduction methods:\n``` python\nfrom kiez import Kiez\nimport numpy as np\n# create example data\nrng = np.random.RandomState(0)\nsource = rng.rand(100,50)\ntarget = rng.rand(100,50)\n# prepare algorithm and hubness reduction\nalgo_kwargs = {"n_candidates": 10}\nk_inst = Kiez(n_neighbors=5, algorithm="Faiss" algorithm_kwargs=algo_kwargs, hubness="CSLS")\n# fit and get neighbors\nk_inst.fit(source, target)\nnn_dist, nn_ind = k_inst.kneighbors()\n```\n\n## Torch Support\nBeginning with version 0.5.0 torch can be used, when using `Faiss` as NN library:\n\n```python\n\n    from kiez import Kiez\n    import torch\n    source = torch.randn((100,10))\n    target = torch.randn((200,10))\n    k_inst = Kiez(algorithm="Faiss", hubness="CSLS")\n    k_inst.fit(source, target)\n    nn_dist, nn_ind = k_inst.kneighbors()\n```\n\nYou can also utilize tensor on the GPU:\n\n```python\n\n    k_inst = Kiez(algorithm="Faiss", algorithm_kwargs={"use_gpu":True}, hubness="CSLS")\n    k_inst.fit(source.cuda(), target.cuda())\n    nn_dist, nn_ind = k_inst.kneighbors()\n```\n\n## Documentation\nYou can find more documentation on [readthedocs](https://kiez.readthedocs.io)\n\n## Benchmark\nThe results and configurations of our experiments can be found in a seperate [benchmarking repository](https://github.com/dobraczka/kiez-benchmarking)\n\n## Citation\nIf you find this work useful you can use the following citation:\n```\n@article{obraczka2022fast,\n  title={Fast Hubness-Reduced Nearest Neighbor Search for Entity Alignment in Knowledge Graphs},\n  author={Obraczka, Daniel and Rahm, Erhard},\n  journal={SN Computer Science},\n  volume={3},\n  number={6},\n  pages={1--19},\n  year={2022},\n  publisher={Springer},\n  url={https://link.springer.com/article/10.1007/s42979-022-01417-1},\n  doi={10.1007/s42979-022-01417-1},\n}\n```\n\n## Contributing\nPRs and enhancement ideas are always welcome. If you want to build kiez locally use:\n```bash\ngit clone git@github.com:dobraczka/kiez.git\ncd kiez\npoetry install\n```\nTo run the tests (given you are in the kiez folder):\n```bash\npoetry run pytest tests\n```\n\nOr install [nox](https://github.com/theacodes/nox) and run:\n```\nnox\n```\nwhich checks all the linting as well.\n\n## License\n`kiez` is licensed under the terms of the BSD-3-Clause [license](LICENSE.txt).\nSeveral files were modified from [`scikit-hubness`](https://github.com/VarIr/scikit-hubness),\ndistributed under the same [license](external/SCIKIT_HUBNESS_LICENSE.txt).\nThe respective files contain the following tag instead of the full license text.\n\n        SPDX-License-Identifier: BSD-3-Clause\n\nThis enables machine processing of license information based on the SPDX\nLicense Identifiers that are here available: https://spdx.org/licenses/\n',
    'author': 'Daniel Obraczka',
    'author_email': 'obraczka@informatik.uni-leipzig.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/dobraczka/kiez',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness: https://github.com/VarIr/scikit-hubness/

from typing import Tuple, TypeVar

import numpy as np
from sklearn.utils.validation import check_is_fitted

from .base import HubnessReduction

T = TypeVar("T")

try:
    import torch
except ImportError:
    torch = None


class LocalScaling(HubnessReduction):
    """Hubness reduction with Local Scaling.

    Uses the formula presented in [1]_.

    Parameters
    ----------
    k: int, default = 5
        Number of neighbors to consider for the rescaling
    method: 'standard' or 'nicdm', default = 'standard'
        Perform local scaling with the specified variant:
        - 'standard' or 'ls' rescale distances using the distance to the k-th neighbor
        - 'nicdm' rescales distances using a statistic over distances to k neighbors
    verbose: int, default = 0
        If verbose > 0, show progress bar.

    References
    ----------
    .. [1] Schnitzer, D., Flexer, A., Schedl, M., & Widmer, G. (2012).
           Local and global scaling reduce hubs in space. The Journal of Machine
           Learning Research, 13(1), 2871-2902.
    """

    def __init__(self, method: str = "standard", **kwargs):
        super().__init__(**kwargs)
        self.method = method.lower()
        if self.method not in ["ls", "standard", "nicdm"]:
            raise ValueError(
                f"Internal: Invalid method {self.method}. Try 'ls' or 'nicdm'."
            )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(method = {self.method}, verbose ="
            f" {self.verbose})"
        )

    def _fit(
        self,
        neigh_dist,
        neigh_ind,
        source,
        target,
    ) -> "LocalScaling":
        """Fit the model using neigh_dist and neigh_ind as training data.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_samples, n_neighbors)
            Distance matrix of training objects (rows) against their
            individual k nearest neighbors (colums).
        neigh_ind: np.ndarray, shape (n_samples, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist.
        source
            Ignored
        target
            Ignored

        Returns
        -------
        LocalScaling
            Fitted LocalScaling
        """
        self.r_dist_t_to_s_ = neigh_dist
        self.r_ind_t_to_s_ = neigh_ind
        return self

    def _exp(self, inner_exp):
        if self._use_torch:
            return torch.exp(inner_exp)
        return np.exp(inner_exp)

    def _sqrt(self, value):
        if self._use_torch:
            return torch.sqrt(value)
        return np.sqrt(value)

    def transform(
        self,
        neigh_dist,
        neigh_ind,
        query=None,
    ) -> Tuple[T, T]:
        """Transform distance between test and training data with Mutual Proximity.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_query, n_neighbors)
            Distance matrix of test objects (rows) against their individual
            k nearest neighbors among the training data (columns).
        neigh_ind: np.ndarray, shape (n_query, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist
        query
            Ignored

        Returns
        -------
        hub_reduced_dist, neigh_ind
            Local scaling distances, and corresponding neighbor indices

        Raises
        ------
        ValueError
            If wrong self.method was supplied

        Notes
        -----
        The returned distances are NOT sorted! If you use this class directly,
        you will need to sort the returned matrices according to hub_reduced_dist.
        """
        check_is_fitted(self, "r_dist_t_to_s_")

        # Find distances to the k-th neighbor (standard LS) or the k neighbors (NICDM)
        r_dist_s_to_t = neigh_dist

        # Perform standard local scaling...
        if self.method in ["ls", "standard"]:
            r_t_to_s = self.r_dist_t_to_s_[:, -1]
            r_s_to_t = r_dist_s_to_t[:, -1].reshape(-1, 1)
            inner_exp = -1 * neigh_dist**2 / (r_s_to_t * r_t_to_s[neigh_ind])
            exp = self._exp(inner_exp)
            hub_reduced_dist = 1.0 - exp
        # ...or use non-iterative contextual dissimilarity measure
        elif self.method == "nicdm":
            r_t_to_s = self.r_dist_t_to_s_.mean(axis=1)
            r_s_to_t = r_dist_s_to_t.mean(axis=1).reshape(-1, 1)
            inner_sqrt = r_s_to_t * r_t_to_s[neigh_ind]
            sqrt = self._sqrt(inner_sqrt)
            hub_reduced_dist = neigh_dist / sqrt

        # Return the hubness reduced distances
        # These must be sorted downstream
        return hub_reduced_dist, neigh_ind
